# Unit tests for flat_review_presenter.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# pylint: disable=wrong-import-position
# pylint: disable=import-outside-toplevel
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-statements
# pylint: disable=protected-access
# pylint: disable=too-many-locals
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments

"""Unit tests for flat_review_presenter.py methods."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from .orca_test_context import OrcaTestContext
    from unittest.mock import MagicMock

@pytest.mark.unit
class TestFlatReviewPresenter:
    """Test FlatReviewPresenter class methods."""

    def _setup_dependencies(self, test_context: OrcaTestContext) -> dict[str, MagicMock]:
        """Set up mocks for flat_review_presenter dependencies."""

        additional_modules = [
            "gi",
            "gi.repository",
            "gi.repository.Atspi",
            "gi.repository.Gtk",
            "gi.repository.GLib",
            "orca.flat_review",
            "orca.speech_and_verbosity_manager",
            "orca.ax_event_synthesizer",
            "orca.ax_text",
        ]
        essential_modules = test_context.setup_shared_dependencies(additional_modules)

        gi_mock = essential_modules["gi"]
        gi_mock.require_version = test_context.Mock()

        gi_repository_mock = essential_modules["gi.repository"]
        atspi_mock = essential_modules["gi.repository.Atspi"]
        gi_repository_mock.Atspi = atspi_mock

        gtk_mock = essential_modules["gi.repository.Gtk"]
        window_mock = test_context.Mock()
        gtk_mock.Window = test_context.Mock(return_value=window_mock)
        gtk_mock.VBox = test_context.Mock()
        gtk_mock.Label = test_context.Mock()
        gtk_mock.Button = test_context.Mock()

        glib_mock = essential_modules["gi.repository.GLib"]
        glib_error_mock = type("GError", (Exception,), {})
        glib_mock.GError = glib_error_mock

        flat_review_mock = essential_modules["orca.flat_review"]
        flat_review_context_mock = test_context.Mock()

        flat_review_context_mock.get_current_object = test_context.Mock(
            return_value=test_context.Mock()
        )
        flat_review_context_mock.get_current_word = test_context.Mock(return_value="test word")
        flat_review_context_mock.get_current_line = test_context.Mock(return_value="test line")
        flat_review_context_mock.get_current_line_string = test_context.Mock(
            return_value="test line"
        )
        flat_review_context_mock.get_current_char = test_context.Mock(return_value="t")
        flat_review_context_mock.get_current_character = test_context.Mock(return_value="t")
        flat_review_context_mock.get_current_item = test_context.Mock(return_value="test item")
        flat_review_context_mock.get_contents = test_context.Mock(return_value="test content")

        flat_review_context_mock.go_next_line = test_context.Mock(return_value=True)
        flat_review_context_mock.go_previous_line = test_context.Mock(return_value=True)
        flat_review_context_mock.go_next_item = test_context.Mock(return_value=True)
        flat_review_context_mock.go_previous_item = test_context.Mock(return_value=True)
        flat_review_context_mock.go_next_character = test_context.Mock(return_value=True)
        flat_review_context_mock.go_previous_character = test_context.Mock(return_value=True)
        flat_review_context_mock.go_begin_line = test_context.Mock(return_value=True)
        flat_review_context_mock.go_end_line = test_context.Mock(return_value=True)
        flat_review_context_mock.go_top_left = test_context.Mock(return_value=True)
        flat_review_context_mock.go_bottom_right = test_context.Mock(return_value=True)
        flat_review_context_mock.go_bottom_left = test_context.Mock(return_value=True)
        flat_review_context_mock.go_to_start_of = test_context.Mock(return_value=True)
        flat_review_context_mock.go_to_end_of = test_context.Mock(return_value=True)

        context_class_mock = test_context.Mock(return_value=flat_review_context_mock)
        context_class_mock.WINDOW = "window"
        context_class_mock.LINE = "line"
        flat_review_mock.Context = context_class_mock

        focus_manager_mock = essential_modules["orca.focus_manager"]
        focus_manager_instance = test_context.Mock()
        focus_manager_instance.get_active_mode_and_object_of_interest = test_context.Mock(
            return_value=("focus_tracking", test_context.Mock())
        )
        focus_manager_instance.get_locus_of_focus = test_context.Mock(
            return_value=test_context.Mock()
        )
        focus_manager_instance.emit_region_changed = test_context.Mock()
        focus_manager_instance.FLAT_REVIEW = "flat_review"
        focus_manager_instance.FOCUS_TRACKING = "focus_tracking"
        focus_manager_mock.get_manager = test_context.Mock(return_value=focus_manager_instance)
        focus_manager_mock.FLAT_REVIEW = "flat_review"
        focus_manager_mock.FOCUS_TRACKING = "focus_tracking"

        script_manager_mock = essential_modules["orca.script_manager"]
        script_manager_instance = test_context.Mock()
        script_instance = test_context.Mock()
        script_instance.present_message = test_context.Mock()
        script_instance.present_object = test_context.Mock()
        script_instance.update_braille = test_context.Mock()
        script_instance.speak_message = test_context.Mock()

        speech_generator_mock = test_context.Mock()
        speech_generator_mock.voice = test_context.Mock(return_value="voice")
        script_instance.speech_generator = speech_generator_mock

        script_manager_instance.get_active_script = test_context.Mock(return_value=script_instance)
        script_manager_mock.get_manager = test_context.Mock(return_value=script_manager_instance)

        dbus_service_mock = essential_modules["orca.dbus_service"]
        controller_mock = test_context.Mock()
        controller_mock.register_decorated_module = test_context.Mock()
        dbus_service_mock.get_remote_controller = test_context.Mock(return_value=controller_mock)
        dbus_service_mock.command = lambda func: func

        settings_manager_mock = essential_modules["orca.settings_manager"]
        settings_manager_instance = test_context.Mock()
        settings_manager_instance.get_setting = test_context.Mock(return_value=False)
        settings_manager_mock.get_manager = test_context.Mock(
            return_value=settings_manager_instance
        )

        settings_mock = essential_modules["orca.settings"]
        settings_mock.VERBOSITY_LEVEL_BRIEF = 0
        settings_mock.speechVerbosityLevel = 1

        keybindings_mock = essential_modules["orca.keybindings"]
        bindings_instance = test_context.Mock()
        bindings_instance.add = test_context.Mock()
        bindings_instance.remove = test_context.Mock()
        keybindings_mock.KeyBindings = test_context.Mock(return_value=bindings_instance)

        input_event_mock = essential_modules["orca.input_event"]
        input_event_handler_mock = test_context.Mock()
        input_event_mock.InputEventHandler = test_context.Mock(
            return_value=input_event_handler_mock
        )
        input_event_mock.InputEvent = test_context.Mock()
        braille_event_mock = type("BrailleEvent", (), {})
        input_event_mock.BrailleEvent = braille_event_mock

        messages_mock = essential_modules["orca.messages"]
        messages_mock.FLAT_REVIEW_START = "Entering flat review."
        messages_mock.FLAT_REVIEW_STOP = "Leaving flat review."
        messages_mock.FLAT_REVIEW_HOME = "Top left."
        messages_mock.FLAT_REVIEW_END = "Bottom right."

        guilabels_mock = essential_modules["orca.guilabels"]
        guilabels_mock.FLAT_REVIEW = "Flat Review"

        cmdnames_mock = essential_modules["orca.cmdnames"]
        cmdnames_mock.TOGGLE_FLAT_REVIEW = "toggleFlatReviewMode"
        cmdnames_mock.FLAT_REVIEW_HOME = "flatReviewHome"
        cmdnames_mock.FLAT_REVIEW_END = "flatReviewEnd"

        braille_mock = essential_modules["orca.braille"]
        braille_mock.refresh = test_context.Mock()
        braille_mock.present_line = test_context.Mock()
        braille_mock.present_item = test_context.Mock()

        debug_mock = essential_modules["orca.debug"]
        debug_mock.print_message = test_context.Mock()
        debug_mock.LEVEL_INFO = 800

        ax_event_synthesizer_mock = essential_modules["orca.ax_event_synthesizer"]
        ax_event_synthesizer_class_mock = test_context.Mock()
        ax_event_synthesizer_mock.AXEventSynthesizer = ax_event_synthesizer_class_mock

        ax_object_mock = essential_modules["orca.ax_object"]
        ax_object_class_mock = test_context.Mock()
        ax_object_class_mock.get_name = test_context.Mock(return_value="Test Object")
        ax_object_class_mock.is_dead = test_context.Mock(return_value=False)
        ax_object_mock.AXObject = ax_object_class_mock

        ax_text_mock = essential_modules["orca.ax_text"]
        ax_text_class_mock = test_context.Mock()
        ax_text_class_mock.get_text = test_context.Mock(return_value="test text")
        ax_text_mock.AXText = ax_text_class_mock

        speech_verbosity_mock = essential_modules["orca.speech_and_verbosity_manager"]
        speech_verbosity_instance = test_context.Mock()
        speech_verbosity_mock.get_manager = test_context.Mock(
            return_value=speech_verbosity_instance
        )

        essential_modules["flat_review_context"] = flat_review_context_mock
        essential_modules["focus_manager_instance"] = focus_manager_instance
        essential_modules["script_manager_instance"] = script_manager_instance
        essential_modules["script_instance"] = script_instance
        essential_modules["controller"] = controller_mock
        essential_modules["window"] = window_mock
        essential_modules["settings_manager_instance"] = settings_manager_instance

        return essential_modules

    def test_init(self, test_context: OrcaTestContext) -> None:
        """Test FlatReviewPresenter.__init__."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        mock_controller = test_context.Mock()
        test_context.patch(
            "orca.flat_review_presenter.dbus_service.get_remote_controller",
            return_value=mock_controller
        )
        presenter = FlatReviewPresenter()

        assert presenter._context is None
        assert presenter._current_contents == ""
        assert isinstance(presenter._restrict, bool)
        assert presenter._handlers is not None
        assert presenter._desktop_bindings is not None
        assert presenter._laptop_bindings is not None
        assert presenter._gui is None

        mock_controller.register_decorated_module.assert_called_with(
            "FlatReviewPresenter", presenter
        )

    @pytest.mark.parametrize(
        "has_context,expected",
        [
            (False, False),
            (True, True),
        ],
    )
    def test_is_active(
        self, test_context: OrcaTestContext, has_context: bool, expected: bool
    ) -> None:
        """Test FlatReviewPresenter.is_active with various context states."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        presenter._context = test_context.Mock() if has_context else None
        result = presenter.is_active()
        assert result is expected

    def test_get_or_create_context_creates_new_unrestricted(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test FlatReviewPresenter.get_or_create_context creates context in unrestricted mode."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        presenter._context = None
        presenter._restrict = False
        script_mock = test_context.Mock()
        context = presenter.get_or_create_context(script_mock)

        assert context is not None
        assert presenter._context is not None
        essential_modules["orca.flat_review"].Context.assert_called_with(script_mock)
        essential_modules["focus_manager_instance"].emit_region_changed.assert_called_once()

    def test_get_or_create_context_creates_new_restricted(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test FlatReviewPresenter.get_or_create_context creates new context in restricted mode."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        presenter._context = None
        presenter._restrict = True
        script_mock = test_context.Mock()
        root_obj = test_context.Mock()
        essential_modules[
            "focus_manager_instance"
        ].get_active_mode_and_object_of_interest.return_value = ("focus_tracking", root_obj)
        context = presenter.get_or_create_context(script_mock)

        assert context is not None
        assert presenter._context is not None
        essential_modules["orca.flat_review"].Context.assert_called_with(script_mock, root=root_obj)

    def test_get_or_create_context_returns_existing(self, test_context: OrcaTestContext) -> None:
        """Test FlatReviewPresenter.get_or_create_context returns existing context."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        existing_context = test_context.Mock()
        presenter._context = existing_context
        presenter._restrict = False
        script_mock = test_context.Mock()
        context = presenter.get_or_create_context(script_mock)

        assert context is existing_context
        assert essential_modules["orca.flat_review"].Context.call_count == 0

    @pytest.mark.parametrize(
        "refresh,is_desktop,expects_setup_call",
        [
            pytest.param(False, True, False, id="no_refresh_desktop"),
            pytest.param(False, False, False, id="no_refresh_laptop"),
            pytest.param(True, True, True, id="refresh_desktop"),
        ],
    )
    def test_get_bindings(
        self,
        test_context: OrcaTestContext,
        refresh: bool,
        is_desktop: bool,
        expects_setup_call: bool,
    ) -> None:
        """Test FlatReviewPresenter.get_bindings with various refresh and desktop mode settings."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()

        if expects_setup_call:
            mock_setup = test_context.patch_object(presenter, "_setup_bindings")

        bindings = presenter.get_bindings(refresh=refresh, is_desktop=is_desktop)

        if is_desktop:
            assert bindings is presenter._desktop_bindings
        else:
            assert bindings is presenter._laptop_bindings

        if expects_setup_call:
            mock_setup.assert_called_once()

    def test_get_braille_bindings(self, test_context: OrcaTestContext) -> None:
        """Test FlatReviewPresenter.get_braille_bindings."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        braille_bindings = presenter.get_braille_bindings()
        assert isinstance(braille_bindings, dict)

    @pytest.mark.parametrize(
        "refresh",
        [
            pytest.param(False, id="cached"),
            pytest.param(True, id="refresh"),
        ],
    )
    def test_get_handlers(self, test_context: OrcaTestContext, refresh: bool) -> None:
        """Test FlatReviewPresenter.get_handlers with refresh parameter."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()

        if refresh:
            mock_setup = test_context.patch_object(presenter, "_setup_handlers")

        handlers = presenter.get_handlers(refresh=refresh)
        assert handlers is presenter._handlers

        if refresh:
            mock_setup.assert_called_once()

    @pytest.mark.parametrize(
        "has_context,provides_script,provides_event,verbose_mode",
        [
            (True, False, False, False),
            (False, True, False, False),
            (False, True, True, True),
        ],
    )
    def test_start_scenarios(
        self,
        test_context: OrcaTestContext,
        has_context: bool,
        provides_script: bool,
        provides_event: bool,
        verbose_mode: bool,
    ) -> None:
        """Test FlatReviewPresenter.start under different scenarios."""
        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        presenter._context = test_context.Mock() if has_context else None
        script_mock = test_context.Mock() if provides_script else None
        event_mock = test_context.Mock() if provides_event else None
        if verbose_mode:
            settings_mock = essential_modules["orca.settings"]
            settings_mock.speechVerbosityLevel = 1
        presenter.start(script=script_mock, event=event_mock)
        if has_context:
            essential_modules["orca.flat_review"].Context.assert_not_called()
        else:
            assert presenter._context is not None
            essential_modules["orca.flat_review"].Context.assert_called_once()

    @pytest.mark.parametrize("is_active", [False, True])
    def test_quit_scenarios(self, test_context: OrcaTestContext, is_active: bool) -> None:
        """Test FlatReviewPresenter.quit when active or inactive."""
        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        presenter._context = test_context.Mock() if is_active else None
        if is_active:
            focus_obj = test_context.Mock()
            essential_modules["focus_manager_instance"].get_locus_of_focus.return_value = focus_obj
        presenter.quit()
        if is_active:
            assert presenter._context is None
            essential_modules["focus_manager_instance"].emit_region_changed.assert_called_with(
                focus_obj, mode=essential_modules["focus_manager_instance"].FOCUS_TRACKING
            )
        else:
            essential_modules["focus_manager_instance"].emit_region_changed.assert_not_called()

    def test_toggle_flat_review_mode_enter(self, test_context: OrcaTestContext) -> None:
        """Test FlatReviewPresenter.toggle_flat_review_mode enters flat review."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        presenter._context = None
        script_mock = test_context.Mock()
        event_mock = test_context.Mock()
        mock_start = test_context.patch_object(presenter, "start")
        presenter.toggle_flat_review_mode(script_mock, event_mock)
        mock_start.assert_called_once_with(script_mock, event_mock)

    def test_toggle_flat_review_mode_exit(self, test_context: OrcaTestContext) -> None:
        """Test FlatReviewPresenter.toggle_flat_review_mode exits flat review."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        presenter._context = test_context.Mock()
        script_mock = test_context.Mock()
        event_mock = test_context.Mock()
        mock_quit = test_context.patch_object(presenter, "quit")
        presenter.toggle_flat_review_mode(script_mock, event_mock)
        mock_quit.assert_called_once_with(script_mock, event_mock)

    @pytest.mark.parametrize(
        "method_name,context_method,navigation_target",
        [
            ("go_home", "go_to_start_of", "WINDOW"),
            ("go_end", "go_to_end_of", "WINDOW"),
        ],
    )
    def test_navigation_methods(
        self,
        test_context: OrcaTestContext,
        method_name: str,
        context_method: str,
        navigation_target: str,
    ) -> None:
        """Test FlatReviewPresenter navigation methods (go_home, go_end)."""
        essential_modules = self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        script_mock = test_context.Mock()
        event_mock = test_context.Mock()
        context_mock = test_context.Mock()
        presenter._context = context_mock
        mock_present_line = test_context.patch_object(presenter, "present_line")
        method = getattr(presenter, method_name)
        result = method(script_mock, event_mock)
        getattr(context_mock, context_method).assert_called_once_with(
            getattr(essential_modules["orca.flat_review"].Context, navigation_target)
        )
        mock_present_line.assert_called_once_with(script_mock, event_mock)
        assert result is True

    def test_go_bottom_left(self, test_context: OrcaTestContext) -> None:
        """Test FlatReviewPresenter.go_bottom_left."""

        essential_modules = self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        script_mock = test_context.Mock()
        event_mock = test_context.Mock()

        context_mock = test_context.Mock()
        presenter._context = context_mock
        mock_present_line = test_context.patch_object(presenter, "present_line")

        result = presenter.go_bottom_left(script_mock, event_mock)

        assert context_mock.go_to_end_of.call_count == 1
        context_mock.go_to_end_of.assert_called_with(
            essential_modules["orca.flat_review"].Context.WINDOW
        )
        context_mock.go_to_start_of.assert_called_once_with(
            essential_modules["orca.flat_review"].Context.LINE
        )
        mock_present_line.assert_called_once_with(script_mock, event_mock)
        assert result is True

    @pytest.mark.parametrize("navigation_succeeds", [True, False])
    def test_go_previous_line_scenarios(
        self, test_context: OrcaTestContext, navigation_succeeds: bool
    ) -> None:
        """Test FlatReviewPresenter.go_previous_line with success and failure."""
        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        script_mock = test_context.Mock()
        event_mock = test_context.Mock()
        context_mock = test_context.Mock()
        context_mock.go_previous_line.return_value = navigation_succeeds
        presenter._context = context_mock
        mock_present_line = test_context.patch_object(presenter, "present_line")
        result = presenter.go_previous_line(script_mock, event_mock)
        context_mock.go_previous_line.assert_called_once()
        if navigation_succeeds:
            mock_present_line.assert_called_once_with(script_mock, event_mock)
        else:
            mock_present_line.assert_not_called()
        assert result is True

    def test_get_presenter_singleton(self, test_context: OrcaTestContext) -> None:
        """Test get_presenter function returns singleton instance."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import get_presenter

        presenter1 = get_presenter()
        presenter2 = get_presenter()
        assert presenter1 is presenter2
        assert presenter1 is not None

    def test_full_navigation_workflow(self, test_context: OrcaTestContext) -> None:
        """Test complete flat review navigation workflow."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        script_mock = test_context.Mock()
        event_mock = test_context.Mock()

        presenter.start(script_mock, event_mock)
        assert presenter.is_active()

        presenter.quit(script_mock, event_mock)
        assert not presenter.is_active()

    def test_restricted_vs_unrestricted_context_creation(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test context creation in both restricted and unrestricted modes."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter_unrestricted = FlatReviewPresenter()
        presenter_unrestricted._restrict = False
        presenter_unrestricted._context = None
        script_mock = test_context.Mock()
        presenter_unrestricted.get_or_create_context(script_mock)
        essential_modules["orca.flat_review"].Context.assert_called_with(script_mock)

        presenter_restricted = FlatReviewPresenter()
        presenter_restricted._restrict = True
        presenter_restricted._context = None
        root_obj = test_context.Mock()
        essential_modules[
            "focus_manager_instance"
        ].get_active_mode_and_object_of_interest.return_value = ("focus_tracking", root_obj)
        presenter_restricted.get_or_create_context(script_mock)

        assert essential_modules["orca.flat_review"].Context.call_count >= 2

    def test_braille_integration(self, test_context: OrcaTestContext) -> None:
        """Test flat review integration with braille system."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        braille_bindings = presenter.get_braille_bindings()

        assert isinstance(braille_bindings, dict)

    def test_keyboard_bindings_setup(self, test_context: OrcaTestContext) -> None:
        """Test keyboard bindings setup for both desktop and laptop modes."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()

        desktop_bindings = presenter.get_bindings(refresh=False, is_desktop=True)
        assert desktop_bindings is presenter._desktop_bindings

        laptop_bindings = presenter.get_bindings(refresh=False, is_desktop=False)
        assert laptop_bindings is presenter._laptop_bindings

        assert desktop_bindings is not None
        assert laptop_bindings is not None

    def test_dbus_command_registration(self, test_context: OrcaTestContext) -> None:
        """Test D-Bus command registration during initialization."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        mock_controller = test_context.Mock()
        test_context.patch(
            "orca.flat_review_presenter.dbus_service.get_remote_controller",
            return_value=mock_controller
        )
        presenter = FlatReviewPresenter()

        mock_controller.register_decorated_module.assert_called_with(
            "FlatReviewPresenter", presenter
        )

    def test_focus_manager_integration(self, test_context: OrcaTestContext) -> None:
        """Test integration with focus manager for region changes."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        focus_manager_instance = essential_modules["focus_manager_instance"]

        script_mock = test_context.Mock()
        presenter.get_or_create_context(script_mock)
        focus_manager_instance.emit_region_changed.assert_called()

        presenter._context = test_context.Mock()
        focus_obj = test_context.Mock()
        focus_manager_instance.get_locus_of_focus.return_value = focus_obj
        presenter.quit()
        focus_manager_instance.emit_region_changed.assert_called_with(
            focus_obj, mode=focus_manager_instance.FOCUS_TRACKING
        )

    def test_error_handling_navigation_failure(self, test_context: OrcaTestContext) -> None:
        """Test error handling when navigation operations fail."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        context_mock = test_context.Mock()
        context_mock.go_next_line.return_value = False
        presenter._context = context_mock
        script_mock = test_context.Mock()
        event_mock = test_context.Mock()

        presenter.go_next_line(script_mock, event_mock)
        context_mock.go_next_line.assert_called_once()

    def test_gui_lifecycle_management(self, test_context: OrcaTestContext) -> None:
        """Test GUI-related methods exist."""

        self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()

        assert hasattr(presenter, "show_contents")
        assert callable(getattr(presenter, "show_contents", None))

    def test_context_caching_behavior(self, test_context: OrcaTestContext) -> None:
        """Test context caching and reuse behavior."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.flat_review_presenter import FlatReviewPresenter

        presenter = FlatReviewPresenter()
        script_mock = test_context.Mock()

        essential_modules["orca.flat_review"].Context.reset_mock()

        context1 = presenter.get_or_create_context(script_mock)
        assert context1 is not None

        context2 = presenter.get_or_create_context(script_mock)
        assert context1 is context2

        assert essential_modules["orca.flat_review"].Context.call_count == 1

        presenter._context = None
        context3 = presenter.get_or_create_context(script_mock)
        assert context3 is not None

        assert essential_modules["orca.flat_review"].Context.call_count == 2



# Unit tests for action_presenter.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=import-outside-toplevel
# pylint: disable=protected-access
# pylint: disable=too-many-public-methods

"""Unit tests for action_presenter.py methods."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from .orca_test_context import OrcaTestContext
    from unittest.mock import MagicMock

LOCATION_NOT_FOUND_FULL_MSG = "Location not found (full)"
LOCATION_NOT_FOUND_BRIEF_MSG = "Location not found (brief)"
LOCATION_NOT_FOUND_MSG = "Location not found"
SHOW_ACTIONS_LIST_CMD = "showActionsList"

@pytest.mark.unit
class TestActionPresenter:
    """Test ActionPresenter class methods."""

    def _setup_dependencies(self, test_context: OrcaTestContext) -> dict[str, MagicMock]:
        """Returns all dependencies needed for ActionPresenter testing."""

        additional_modules = ["orca.brltablenames", "orca.speech"]
        essential_modules = test_context.setup_shared_dependencies(additional_modules)

        essential_modules["orca.cmdnames"].SHOW_ACTIONS_LIST = SHOW_ACTIONS_LIST_CMD
        focus_manager_instance = test_context.Mock()
        focus_manager_instance.get_active_mode_and_object_of_interest = test_context.Mock(
            return_value=(None, None)
        )
        focus_manager_instance.get_locus_of_focus = test_context.Mock(return_value=None)
        focus_manager_instance.get_active_window = test_context.Mock(return_value=None)
        focus_manager_instance.clear_state = test_context.Mock()
        focus_manager_instance.set_active_window = test_context.Mock()
        focus_manager_instance.set_locus_of_focus = test_context.Mock()
        essential_modules["orca.focus_manager"].get_manager = test_context.Mock(
            return_value=focus_manager_instance
        )

        input_event_handler_mock = test_context.Mock()
        essential_modules["orca.input_event"].InputEventHandler = test_context.Mock(
            return_value=input_event_handler_mock
        )
        essential_modules["orca.input_event"].InputEvent = test_context.mocker.MagicMock

        script_manager_instance = essential_modules["orca.script_manager"].get_manager()
        script_mock = script_manager_instance.get_active_script()
        script_mock.speech_generator.get_localized_role_name = test_context.Mock(
            return_value="button"
        )

        essential_modules["orca.messages"].LOCATION_NOT_FOUND_FULL = LOCATION_NOT_FOUND_FULL_MSG
        essential_modules["orca.messages"].LOCATION_NOT_FOUND_BRIEF = LOCATION_NOT_FOUND_BRIEF_MSG
        essential_modules["orca.messages"].NO_ACTIONS_FOUND_ON = "No actions found on %s"
        essential_modules["orca.guilabels"].ACTIONS_LIST = "Actions List"
        return essential_modules

    # pylint: disable-next=too-many-locals
    @pytest.mark.parametrize(
        "test_scenario,n_actions,action_descriptions,object_source,"
        "expected_message,expects_gui_creation",
        [
            pytest.param(
                "with_actions",
                2,
                ["Click button", "Activate button"],
                "locus_of_focus",
                None,
                True,
                id="actions_from_locus",
            ),
            pytest.param(
                "no_object",
                0,
                [],
                None,
                LOCATION_NOT_FOUND_FULL_MSG,
                False,
                id="no_object_available",
            ),
            pytest.param(
                "empty_descriptions",
                2,
                ["", ""],
                "locus_of_focus",
                None,
                True,
                id="empty_action_descriptions",
            ),
            pytest.param(
                "from_active_mode",
                2,
                ["Click button", "Activate button"],
                "active_mode",
                None,
                True,
                id="object_from_active_mode",
            ),
            pytest.param(
                "no_actions_found",
                0,
                [],
                "locus_of_focus",
                "No actions found on Test Button",
                False,
                id="object_has_no_actions",
            ),
        ],
    )
    def test_show_actions_list(  # pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-locals,too-many-statements
        self,
        test_context: OrcaTestContext,
        test_scenario: str,
        n_actions: int,
        action_descriptions: list[str],
        object_source: str | None,
        expected_message: str | None,
        expects_gui_creation: bool,
    ) -> None:
        """Test ActionPresenter.show_actions_list with various scenarios."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)

        if test_scenario != "no_object":
            ax_object_mock = test_context.Mock()
            ax_object_mock.get_n_actions = test_context.Mock(return_value=n_actions)
            ax_object_mock.get_action_name = test_context.Mock(
                side_effect=lambda obj, i: ["click", "activate"][i]
                if i < len(["click", "activate"])
                else f"action{i}"
            )
            ax_object_mock.get_action_description = test_context.Mock(
                side_effect=lambda obj, i: action_descriptions[i]
                if i < len(action_descriptions)
                else f"desc{i}"
            )
            ax_object_mock.get_name = test_context.Mock(return_value="Test Button")

            ax_utilities_mock = test_context.Mock()
            ax_utilities_mock.get_application = test_context.Mock(return_value=test_context.Mock())

            ax_object_module_mock = test_context.Mock()
            ax_object_module_mock.AXObject = ax_object_mock
            test_context.patch_module("orca.ax_object", ax_object_module_mock)
            ax_utilities_module_mock = test_context.Mock()
            ax_utilities_module_mock.AXUtilities = ax_utilities_mock
            test_context.patch_module("orca.ax_utilities", ax_utilities_module_mock)

        from orca.action_presenter import ActionPresenter

        presenter = ActionPresenter()
        mock_script = essential_modules["orca.script_manager"].get_manager().get_script()
        mock_event = test_context.Mock()
        focus_manager = essential_modules["orca.focus_manager"].get_manager()

        if object_source == "locus_of_focus":
            mock_accessible_object = test_context.Mock()
            mock_accessible_object.get_name = test_context.Mock(return_value="Test Button")
            focus_manager.get_active_mode_and_object_of_interest.return_value = (None, None)
            focus_manager.get_locus_of_focus.return_value = mock_accessible_object
        elif object_source == "active_mode":
            mock_accessible_object = test_context.Mock()
            mock_accessible_object.get_name = test_context.Mock(return_value="Test Button")
            focus_manager.get_active_mode_and_object_of_interest.return_value = (
                test_context.Mock(),
                mock_accessible_object,
            )
            focus_manager.get_locus_of_focus.return_value = None
        else:
            focus_manager.get_active_mode_and_object_of_interest.return_value = (None, None)
            focus_manager.get_locus_of_focus.return_value = None
            mock_accessible_object = None

        mock_action_list_class = None
        if expects_gui_creation:
            mock_window = test_context.Mock()
            focus_manager.get_active_window.return_value = mock_window
            mock_action_list = test_context.Mock()
            mock_action_list_class = test_context.patch(
                "orca.action_presenter.ActionList", return_value=mock_action_list
            )

        result = presenter.show_actions_list(mock_script, mock_event, True)
        assert result is True

        if expected_message:
            if LOCATION_NOT_FOUND_MSG in expected_message:
                mock_script.present_message.assert_called_once_with(
                    LOCATION_NOT_FOUND_FULL_MSG, LOCATION_NOT_FOUND_BRIEF_MSG
                )
            else:
                mock_script.present_message.assert_called_once_with(expected_message)

        if expects_gui_creation and n_actions > 0:
            assert mock_action_list_class is not None
            mock_action_list_class.assert_called_once()
            mock_action_list.show_gui.assert_called_once()
            assert presenter._obj == mock_accessible_object
            assert presenter._gui == mock_action_list

            if test_scenario == "empty_descriptions":
                call_args = mock_action_list_class.call_args[0][0]
                assert "click" in call_args
                assert call_args["click"] == "click"

    @pytest.mark.parametrize(
        "action_success,gui_state",
        [
            pytest.param(True, "present", id="successful_action"),
            pytest.param(False, "present", id="failed_action"),
            pytest.param(True, None, id="no_gui_early_return"),
        ],
    )
    def test_perform_action(
        self,
        test_context: OrcaTestContext,
        action_success: bool,
        gui_state: str | None,
    ) -> None:
        """Test ActionPresenter._perform_action with various scenarios."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.action_presenter import ActionPresenter

        presenter = ActionPresenter()
        mock_accessible_object = test_context.Mock()
        mock_accessible_object.get_name = test_context.Mock(return_value="Test Button")
        presenter._obj = mock_accessible_object

        if gui_state == "present":
            presenter._gui = test_context.Mock()
        else:
            presenter._gui = None

        mock_idle_add = test_context.patch("orca.action_presenter.GLib.idle_add")
        mock_debug_tokens = test_context.patch("orca.action_presenter.debug.print_tokens")
        mock_do_action = test_context.patch("orca.action_presenter.AXObject.do_named_action")
        mock_do_action.return_value = action_success

        presenter._perform_action("click")

        if gui_state is None:
            essential_modules["orca.debug"].print_message.assert_called()
            debug_call = essential_modules["orca.debug"].print_message.call_args[0][1]
            assert "_perform_action called when self._gui is None" in debug_call
            mock_do_action.assert_not_called()
        else:
            assert presenter._gui is not None
            presenter._gui.hide.assert_called_once()
            mock_do_action.assert_called_with(mock_accessible_object, "click")
            mock_idle_add.assert_called_with(presenter._gui.destroy)
            mock_debug_tokens.assert_called()

            if not action_success:
                debug_calls = mock_debug_tokens.call_args_list
                assert len(debug_calls) > 0

    def test_get_presenter_singleton(self, test_context: OrcaTestContext) -> None:
        """Test action_presenter.get_presenter singleton functionality."""

        self._setup_dependencies(test_context)
        from orca.action_presenter import get_presenter

        presenter1 = get_presenter()
        presenter2 = get_presenter()
        assert presenter1 is presenter2
        assert presenter1.__class__.__name__ == "ActionPresenter"

    @pytest.mark.parametrize(
        "method_name,refresh,mock_empty,expects_debug",
        [
            pytest.param("get_bindings", True, False, True, id="bindings_with_refresh"),
            pytest.param("get_bindings", False, True, False, id="bindings_empty_no_refresh"),
            pytest.param("get_handlers", True, False, True, id="handlers_with_refresh"),
        ],
    )
    def test_get_bindings_and_handlers(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        test_context: OrcaTestContext,
        method_name: str,
        refresh: bool,
        mock_empty: bool,
        expects_debug: bool,
    ) -> None:
        """Test ActionPresenter.get_bindings and get_handlers with various settings."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.action_presenter import ActionPresenter

        presenter = ActionPresenter()

        if method_name == "get_bindings":
            mock_setup = test_context.patch_object(presenter, "_setup_bindings", return_value=None)

            if mock_empty:
                test_context.patch_object(
                    presenter._bindings, "is_empty", return_value=True
                )

            bindings_result = presenter.get_bindings(refresh=refresh, is_desktop=bool(refresh))
            assert bindings_result is presenter._bindings
        else:
            mock_setup = test_context.patch_object(presenter, "_setup_handlers", return_value=None)
            handlers_result = presenter.get_handlers(refresh=refresh)
            assert handlers_result is presenter._handlers

        mock_setup.assert_called_once()
        if expects_debug:
            essential_modules["orca.debug"].print_message.assert_called()

    def test_setup_bindings_creates_key_binding(self, test_context: OrcaTestContext) -> None:
        """Test ActionPresenter._setup_bindings creates proper key binding."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.action_presenter import ActionPresenter

        presenter = ActionPresenter()

        presenter._bindings = test_context.Mock()
        presenter._bindings.is_empty = test_context.Mock(return_value=True)

        mock_keybindings = test_context.Mock()
        mock_key_binding_class = test_context.Mock()
        mock_keybindings.KeyBindings = test_context.Mock()
        mock_keybindings.KeyBinding = mock_key_binding_class
        mock_keybindings.DEFAULT_MODIFIER_MASK = "DEFAULT_MODIFIER"
        mock_keybindings.ORCA_SHIFT_MODIFIER_MASK = "ORCA_SHIFT_MODIFIER"
        test_context.patch("orca.action_presenter.keybindings", new=mock_keybindings)

        presenter._setup_bindings()

        mock_key_binding_class.assert_called_once_with(
            "a", "DEFAULT_MODIFIER", "ORCA_SHIFT_MODIFIER", presenter._handlers["show_actions_list"]
        )
        essential_modules["orca.debug"].print_message.assert_called()

    def test_restore_focus_sets_script_and_focus(self, test_context: OrcaTestContext) -> None:
        """Test ActionPresenter._restore_focus properly restores script and focus."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.action_presenter import ActionPresenter

        presenter = ActionPresenter()
        mock_obj = test_context.Mock()
        mock_window = test_context.Mock()
        mock_app = test_context.Mock()
        mock_script = test_context.Mock()

        presenter._obj = mock_obj
        presenter._window = mock_window

        ax_utilities_mock = test_context.Mock()
        ax_utilities_mock.get_application = test_context.Mock(return_value=mock_app)
        test_context.patch("orca.action_presenter.AXUtilities", new=ax_utilities_mock)

        script_manager_instance = essential_modules["orca.script_manager"].get_manager()
        script_manager_instance.get_script.return_value = mock_script

        focus_manager_instance = essential_modules["orca.focus_manager"].get_manager()

        presenter._restore_focus()

        ax_utilities_mock.get_application.assert_called_once_with(mock_obj)
        script_manager_instance.get_script.assert_called_once_with(mock_app, mock_obj)
        script_manager_instance.set_active_script.assert_called_once_with(
            mock_script, "Action Presenter list is being destroyed"
        )
        focus_manager_instance.clear_state.assert_called_once_with(
            "Action Presenter list is being destroyed"
        )
        focus_manager_instance.set_active_window.assert_called_once_with(mock_window)
        focus_manager_instance.set_locus_of_focus.assert_called_once_with(None, mock_obj)

    def test_clear_gui_and_restore_focus(self, test_context: OrcaTestContext) -> None:
        """Test ActionPresenter._clear_gui_and_restore_focus clears gui and calls restore."""

        self._setup_dependencies(test_context)
        from orca.action_presenter import ActionPresenter

        presenter = ActionPresenter()
        presenter._gui = test_context.Mock()

        mock_restore_focus = test_context.patch_object(
            presenter, "_restore_focus", return_value=None
        )

        presenter._clear_gui_and_restore_focus()

        assert presenter._gui is None
        mock_restore_focus.assert_called_once()



# Unit tests for object_navigator.py methods.
#
# Copyright 2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=wrong-import-position
# pylint: disable=import-outside-toplevel
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-statements
# pylint: disable=protected-access
# pylint: disable=too-many-locals

"""Unit tests for object_navigator.py methods."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from .orca_test_context import OrcaTestContext
    from unittest.mock import MagicMock

@pytest.mark.unit
class TestObjectNavigator:
    """Test ObjectNavigator class methods."""

    def _setup_dependencies(self, test_context: OrcaTestContext) -> dict[str, MagicMock]:
        """Set up mocks for object_navigator dependencies."""

        import sys

        essential_modules = test_context._setup_essential_modules(
            [
                "gi",
                "gi.repository",
                "gi.repository.Atspi",
                "orca.cmdnames",
                "orca.dbus_service",
                "orca.debug",
                "orca.focus_manager",
                "orca.input_event",
                "orca.keybindings",
                "orca.messages",
                "orca.ax_event_synthesizer",
                "orca.ax_object",
                "orca.ax_utilities",
            ]
        )

        atspi_mock = essential_modules["gi.repository.Atspi"]
        atspi_accessible_mock = test_context.Mock()
        atspi_accessible_mock.get_parent = test_context.Mock(return_value=None)
        atspi_accessible_mock.get_child_count = test_context.Mock(return_value=0)
        atspi_accessible_mock.get_child_at_index = test_context.Mock(return_value=None)
        atspi_mock.Accessible = atspi_accessible_mock

        dbus_service_mock = essential_modules["orca.dbus_service"]
        controller_mock = test_context.Mock()
        controller_mock.register_decorated_module = test_context.Mock()
        controller_mock.reset_mock()
        dbus_service_mock.get_remote_controller = test_context.Mock(return_value=controller_mock)

        def mock_command_decorator(func):
            return func

        dbus_service_mock.command = mock_command_decorator

        if "orca.object_navigator" in sys.modules:
            object_navigator_module = sys.modules["orca.object_navigator"]
            if not hasattr(object_navigator_module, "_original_navigator"):
                setattr(
                    object_navigator_module,
                    "_original_navigator",
                    getattr(object_navigator_module, "_navigator", None),
                )
            setattr(object_navigator_module, "_navigator", test_context.Mock())

        debug_mock = essential_modules["orca.debug"]
        debug_mock.LEVEL_INFO = 800
        debug_mock.print_message = test_context.Mock()
        debug_mock.print_tokens = test_context.Mock()

        input_event_mock = essential_modules["orca.input_event"]
        input_event_handler_mock = test_context.Mock()
        input_event_mock.InputEventHandler = test_context.Mock(
            return_value=input_event_handler_mock
        )

        keybindings_mock = essential_modules["orca.keybindings"]
        keybindings_instance_mock = test_context.Mock()
        keybindings_instance_mock.is_empty = test_context.Mock(return_value=True)
        keybindings_instance_mock.remove_key_grabs = test_context.Mock()
        keybindings_mock.KeyBindings = test_context.Mock(return_value=keybindings_instance_mock)

        ax_object_mock = essential_modules["orca.ax_object"]
        ax_object_mock.AXObject = test_context.Mock()
        ax_object_mock.AXObject.get_parent = test_context.Mock(return_value=None)
        ax_object_mock.AXObject.get_child_count = test_context.Mock(return_value=0)
        ax_object_mock.AXObject.iter_children = test_context.Mock(return_value=[])

        ax_utilities_mock = essential_modules["orca.ax_utilities"]
        ax_utilities_mock.AXUtilities = test_context.Mock()
        ax_utilities_mock.AXUtilities.is_focused = test_context.Mock(return_value=False)
        ax_utilities_mock.AXUtilities.is_paragraph = test_context.Mock(return_value=True)
        ax_utilities_mock.AXUtilities.is_layout_only = test_context.Mock(return_value=False)

        focus_manager_mock = essential_modules["orca.focus_manager"]
        focus_manager_mock.get_manager = test_context.Mock()
        focus_manager_instance = test_context.Mock()
        focus_manager_instance.get_locus_of_focus = test_context.Mock(return_value=None)
        focus_manager_instance.get_active_mode_and_object_of_interest = test_context.Mock(
            return_value=("default", None)
        )
        focus_manager_mock.get_manager.return_value = focus_manager_instance

        ax_event_synthesizer_mock = essential_modules["orca.ax_event_synthesizer"]
        ax_event_synthesizer_mock.AXEventSynthesizer = test_context.Mock()
        ax_event_synthesizer_mock.AXEventSynthesizer.try_all_clickable_actions = test_context.Mock(
            return_value=False
        )
        ax_event_synthesizer_mock.AXEventSynthesizer.click_object = test_context.Mock()

        cmdnames_mock = essential_modules["orca.cmdnames"]
        cmdnames_mock.OBJECT_NAVIGATOR_PARENT = "objectNavigatorParent"
        cmdnames_mock.OBJECT_NAVIGATOR_FIRST_CHILD = "objectNavigatorFirstChild"
        cmdnames_mock.OBJECT_NAVIGATOR_NEXT_SIBLING = "objectNavigatorNextSibling"
        cmdnames_mock.OBJECT_NAVIGATOR_PREVIOUS_SIBLING = "objectNavigatorPreviousSibling"
        cmdnames_mock.OBJECT_NAVIGATOR_TOGGLE_SIMPLIFY = "objectNavigatorToggleSimplify"
        cmdnames_mock.OBJECT_NAVIGATOR_PERFORM_ACTION = "objectNavigatorPerformAction"

        messages_mock = essential_modules["orca.messages"]
        messages_mock.OBJECT_NAVIGATOR_NO_PARENT = "No parent"
        messages_mock.OBJECT_NAVIGATOR_NO_CHILD = "No child"
        messages_mock.OBJECT_NAVIGATOR_NO_SIBLING = "No sibling"
        messages_mock.OBJECT_NAVIGATOR_SIMPLIFY_ON = "Simplify navigation on"
        messages_mock.OBJECT_NAVIGATOR_SIMPLIFY_OFF = "Simplify navigation off"

        atspi_mock.ModifierType = test_context.Mock()
        atspi_mock.ModifierType.ALT = 3
        atspi_mock.ModifierType.CONTROL = 2
        atspi_mock.ModifierType.SHIFT = 0

        return essential_modules

    def test_init_creates_navigator_with_defaults(self, test_context: OrcaTestContext) -> None:
        """Test ObjectNavigator initialization with default settings."""

        self._setup_dependencies(test_context)
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator: ObjectNavigator = ObjectNavigator()
        assert navigator._navigator_focus is None
        assert navigator._last_navigator_focus is None
        assert navigator._last_locus_of_focus is None
        assert navigator._simplify is True
        assert navigator._handlers is not None
        assert navigator._bindings is not None

    def test_init_registers_dbus_commands(self, test_context: OrcaTestContext) -> None:
        """Test ObjectNavigator registers D-Bus commands during initialization."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        dbus_service = essential_modules["orca.dbus_service"]
        controller = dbus_service.get_remote_controller.return_value
        controller.register_decorated_module.reset_mock()
        navigator = ObjectNavigator()
        controller.register_decorated_module.assert_called_once_with("ObjectNavigator", navigator)

    def test_get_bindings_returns_existing_when_not_empty(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test get_bindings returns existing bindings when not empty."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        keybindings_instance = essential_modules["orca.keybindings"].KeyBindings.return_value
        keybindings_instance.is_empty.return_value = False
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        bindings = navigator.get_bindings()
        assert bindings is not None
        assert bindings == navigator._bindings

    def test_get_bindings_creates_new_when_empty(self, test_context: OrcaTestContext) -> None:
        """Test get_bindings creates new bindings when current bindings are empty."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        bindings = navigator.get_bindings()
        assert bindings is not None
        assert essential_modules["orca.keybindings"].KeyBindings.call_count >= 2

    def test_get_bindings_refresh_removes_grabs_and_recreates(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test get_bindings with refresh removes key grabs and recreates bindings."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        keybindings_mock = essential_modules["orca.keybindings"].KeyBindings.return_value
        bindings = navigator.get_bindings(refresh=True)
        keybindings_mock.remove_key_grabs.assert_called_once()
        assert bindings is not None

    @pytest.mark.parametrize(
        "refresh",
        [
            pytest.param(False, id="no_refresh"),
            pytest.param(True, id="refresh"),
        ],
    )
    def test_get_handlers(self, test_context: OrcaTestContext, refresh: bool) -> None:
        """Test get_handlers with refresh parameter."""

        self._setup_dependencies(test_context)
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        original_handlers = navigator._handlers if not refresh else None
        handlers = navigator.get_handlers(refresh=refresh)
        assert handlers is not None
        assert isinstance(handlers, dict)

        if not refresh:
            assert handlers == original_handlers

    def test_include_in_simple_navigation_basic_object(self, test_context: OrcaTestContext) -> None:
        """Test _include_in_simple_navigation with basic accessible object."""

        self._setup_dependencies(test_context)
        mock_obj = test_context.Mock()
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        result = navigator._include_in_simple_navigation(mock_obj)
        assert result is True

    def test_exclude_from_simple_navigation_with_script(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test _exclude_from_simple_navigation with script parameter."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_obj = test_context.Mock()
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        result = navigator._exclude_from_simple_navigation(mock_script, mock_obj)
        assert result is False

    def test_children_returns_list_for_object_with_children(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test _children returns list of child objects."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_obj = test_context.Mock()
        mock_child = test_context.Mock()
        ax_object = essential_modules["orca.ax_object"].AXObject
        ax_object.get_child_count.return_value = 1
        ax_object.iter_children.return_value = [mock_child]
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        children = navigator._children(mock_script, mock_obj)
        assert isinstance(children, list)
        assert len(children) == 1
        assert children[0] == mock_child

    def test_children_returns_empty_list_for_object_without_children(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test _children returns empty list for object with no children."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_obj = test_context.Mock()
        ax_object = essential_modules["orca.ax_object"].AXObject
        ax_object.get_child_count.return_value = 0
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        children = navigator._children(mock_script, mock_obj)
        assert isinstance(children, list)
        assert len(children) == 0

    def test_parent_returns_parent_object(self, test_context: OrcaTestContext) -> None:
        """Test _parent returns parent of accessible object."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_obj = test_context.Mock()
        mock_parent = test_context.Mock()
        ax_object = essential_modules["orca.ax_object"].AXObject
        ax_object.get_parent.return_value = mock_parent
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        parent = navigator._parent(mock_script, mock_obj)
        assert parent == mock_parent
        ax_object.get_parent.assert_called_once_with(mock_obj)

    def test_parent_returns_none_when_no_parent(self, test_context: OrcaTestContext) -> None:
        """Test _parent returns None when object has no parent."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_obj = test_context.Mock()
        ax_object = essential_modules["orca.ax_object"].AXObject
        ax_object.get_parent.return_value = None
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        parent = navigator._parent(mock_script, mock_obj)
        assert parent is None

    def test_set_navigator_focus_updates_focus_tracking(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test _set_navigator_focus properly tracks focus changes."""

        self._setup_dependencies(test_context)
        mock_obj = test_context.Mock()
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._set_navigator_focus(mock_obj)
        assert navigator._navigator_focus == mock_obj

    def test_set_navigator_focus_tracks_previous_focus(self, test_context: OrcaTestContext) -> None:
        """Test _set_navigator_focus tracks previous focus object."""

        self._setup_dependencies(test_context)
        mock_obj1 = test_context.Mock()
        mock_obj2 = test_context.Mock()
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._set_navigator_focus(mock_obj1)
        navigator._set_navigator_focus(mock_obj2)
        assert navigator._navigator_focus == mock_obj2
        assert navigator._last_navigator_focus == mock_obj1

    def test_update_tracks_locus_of_focus(self, test_context: OrcaTestContext) -> None:
        """Test _update tracks the current locus of focus."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        mock_focus_obj = test_context.Mock()
        focus_manager = essential_modules["orca.focus_manager"].get_manager.return_value
        focus_manager.get_locus_of_focus.return_value = mock_focus_obj
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._update()
        assert navigator._last_locus_of_focus == mock_focus_obj

    def test_present_calls_script_presentation_method(self, test_context: OrcaTestContext) -> None:
        """Test _present calls appropriate script method for object presentation."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_script.present_object = test_context.Mock()
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._navigator_focus = test_context.Mock()
        navigator._present(mock_script)
        mock_script.present_object.assert_called_once()

    @pytest.mark.parametrize(
        "notify_user",
        [
            pytest.param(True, id="notify_user_true"),
            pytest.param(False, id="notify_user_false"),
        ],
    )
    def test_present_respects_notify_user_parameter(
        self, test_context: OrcaTestContext, notify_user
    ) -> None:
        """Test _present respects the notify_user parameter."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_script.present_object = test_context.Mock()
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._navigator_focus = test_context.Mock()
        navigator._present(mock_script, notify_user=notify_user)
        if notify_user:
            mock_script.present_object.assert_called_once()
        else:
            mock_script.present_object.assert_not_called()

    @pytest.mark.parametrize(
        "has_parent",
        [
            pytest.param(True, id="success"),
            pytest.param(False, id="no_parent"),
        ],
    )
    def test_move_to_parent_scenarios(
        self, test_context: OrcaTestContext, has_parent: bool
    ) -> None:
        """Test move_to_parent with various parent availability scenarios."""
        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_script.present_object = test_context.Mock()
        mock_current = test_context.Mock()

        if has_parent:
            mock_parent = test_context.Mock()
            ax_object = essential_modules["orca.ax_object"].AXObject
            ax_object.get_parent.return_value = mock_parent
            expected_focus = mock_parent
        else:
            ax_object = essential_modules["orca.ax_object"].AXObject
            ax_object.get_parent.return_value = None
            expected_focus = mock_current

        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._navigator_focus = mock_current
        result = navigator.move_to_parent(mock_script)
        assert result is True
        assert navigator._navigator_focus == expected_focus

    @pytest.mark.parametrize(
        "has_children, child_count",
        [
            pytest.param(True, 1, id="success"),
            pytest.param(False, 0, id="no_children"),
        ],
    )
    def test_move_to_first_child_scenarios(
        self, test_context: OrcaTestContext, has_children: bool, child_count: int
    ) -> None:
        """Test move_to_first_child with various child availability scenarios."""
        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_script.present_object = test_context.Mock()
        mock_current = test_context.Mock()
        ax_object = essential_modules["orca.ax_object"].AXObject
        ax_object.get_child_count.return_value = child_count

        if has_children:
            mock_child = test_context.Mock()
            ax_object.iter_children.return_value = [mock_child]
            expected_focus = mock_child
        else:
            expected_focus = mock_current

        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._navigator_focus = mock_current
        result = navigator.move_to_first_child(mock_script)
        assert result is True
        assert navigator._navigator_focus == expected_focus

    def test_move_to_next_sibling_success(self, test_context: OrcaTestContext) -> None:
        """Test move_to_next_sibling successfully navigates to next sibling."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_script.present_object = test_context.Mock()
        mock_current = test_context.Mock()
        mock_parent = test_context.Mock()
        mock_sibling = test_context.Mock()
        ax_object = essential_modules["orca.ax_object"].AXObject
        ax_object.get_parent.return_value = mock_parent
        ax_object.get_child_count.return_value = 2
        ax_object.iter_children.return_value = [mock_current, mock_sibling]
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._navigator_focus = mock_current
        result = navigator.move_to_next_sibling(mock_script)
        assert result is True
        assert navigator._navigator_focus == mock_sibling

    def test_move_to_previous_sibling_success(self, test_context: OrcaTestContext) -> None:
        """Test move_to_previous_sibling successfully navigates to previous sibling."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_script.present_object = test_context.Mock()
        mock_current = test_context.Mock()
        mock_parent = test_context.Mock()
        mock_sibling = test_context.Mock()
        ax_object = essential_modules["orca.ax_object"].AXObject
        ax_object.get_parent.return_value = mock_parent
        ax_object.get_child_count.return_value = 2
        ax_object.iter_children.return_value = [mock_sibling, mock_current]
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._navigator_focus = mock_current
        result = navigator.move_to_previous_sibling(mock_script)
        assert result is True
        assert navigator._navigator_focus == mock_sibling

    def test_toggle_simplify_changes_simplify_state(self, test_context: OrcaTestContext) -> None:
        """Test toggle_simplify changes the simplification state."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        original_state = navigator._simplify
        result = navigator.toggle_simplify(mock_script)
        assert result is True
        assert navigator._simplify != original_state

    def test_toggle_simplify_announces_state_change(self, test_context: OrcaTestContext) -> None:
        """Test toggle_simplify announces the state change to user."""

        self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_script.present_message = test_context.Mock()
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator.toggle_simplify(mock_script)
        mock_script.present_message.assert_called_once()

    def test_perform_action_calls_synthesizer(self, test_context: OrcaTestContext) -> None:
        """Test perform_action uses AXEventSynthesizer for action execution."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        mock_script = test_context.Mock()
        mock_obj = test_context.Mock()
        ax_event_synthesizer = essential_modules["orca.ax_event_synthesizer"].AXEventSynthesizer
        ax_event_synthesizer.try_all_clickable_actions = test_context.Mock(return_value=False)
        ax_event_synthesizer.click_object = test_context.Mock(return_value=True)
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._navigator_focus = mock_obj
        result = navigator.perform_action(mock_script)
        assert result is True
        ax_event_synthesizer.click_object.assert_called_once_with(mock_obj, 1)

    def test_children_with_null_object(self, test_context: OrcaTestContext) -> None:
        """Test _children handles None object gracefully."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        mock_script = test_context.Mock()
        children = navigator._children(mock_script, None)
        assert isinstance(children, list)
        assert len(children) == 0
        essential_modules["orca.debug"].print_message.assert_called()

    def test_parent_with_null_object_returns_none(self, test_context: OrcaTestContext) -> None:
        """Test _parent returns None when given None object."""

        self._setup_dependencies(test_context)
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        mock_script = test_context.Mock()
        parent = navigator._parent(mock_script, None)
        assert parent is None

    def test_move_to_parent_with_null_navigator_focus(self, test_context: OrcaTestContext) -> None:
        """Test move_to_parent handles None navigator focus gracefully."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._navigator_focus = None
        mock_script = test_context.Mock()
        result = navigator.move_to_parent(mock_script)
        assert result is True
        essential_modules["orca.debug"].print_message.assert_called()

    def test_move_to_first_child_with_dead_object(self, test_context: OrcaTestContext) -> None:
        """Test move_to_first_child handles dead AT-SPI objects."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        mock_script = test_context.Mock()
        mock_obj = test_context.Mock()
        navigator._navigator_focus = mock_obj

        ax_object = essential_modules["orca.ax_object"].AXObject
        ax_object.is_dead.return_value = True

        result = navigator.move_to_first_child(mock_script)
        assert result is True
        essential_modules["orca.debug"].print_message.assert_called()

    def test_perform_action_with_null_focus_logs_debug_info(
        self, test_context: OrcaTestContext
    ) -> None:
        """Test perform_action logs debug information with None navigator focus."""

        essential_modules: dict[str, MagicMock] = self._setup_dependencies(test_context)
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._navigator_focus = None
        mock_script = test_context.Mock()

        result = navigator.perform_action(mock_script)
        assert result is True
        essential_modules["orca.debug"].print_tokens.assert_called()

    def test_toggle_simplify_announces_on_state(self, test_context: OrcaTestContext) -> None:
        """Test toggle_simplify announces state change correctly."""

        self._setup_dependencies(test_context)
        from orca.object_navigator import ObjectNavigator  # pylint: disable=import-outside-toplevel

        navigator = ObjectNavigator()
        navigator._simplify = False
        mock_script = test_context.Mock()
        mock_script.present_message = test_context.Mock()

        navigator.toggle_simplify(mock_script)

        mock_script.present_message.assert_called_once()
        assert navigator._simplify is True

    def test_get_navigator_returns_singleton_instance(self, test_context: OrcaTestContext) -> None:
        """Test get_navigator function returns singleton ObjectNavigator instance."""

        self._setup_dependencies(test_context)
        from orca.object_navigator import get_navigator  # pylint: disable=import-outside-toplevel

        navigator1 = get_navigator()
        navigator2 = get_navigator()
        assert navigator1 is navigator2
        assert navigator1.__class__.__name__ == "ObjectNavigator"



# Orca
#
# Copyright 2006-2008 Sun Microsystems Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Exposes a dictionary, pronunciation_dict, that maps words to what they sound like."""

__id__        = "$Id$"
__version__   = "$Revision$"
__date__      = "$Date$"
__copyright__ = "Copyright (c) 2006-2008 Sun Microsystems Inc."
__license__   = "LGPL"

def get_pronunciation(word: str, pronunciations: dict[str, list[str]] | None = None) -> str:
    """Returns a user-provided string that represents what the word sounds like."""

    dictionary = pronunciations or pronunciation_dict
    entry = dictionary.get(word.lower(), [word, word])
    return entry[1]

def set_pronunciation(
  word: str,
  replacement: str,
  pronunciations: dict[str, list[str]] | None = None
) -> None:
    """Adds word/replacement pair to the pronunciation dictionary."""

    key = word.lower()
    if pronunciations is not None:
        pronunciations[key] = [word, replacement]
    else:
        pronunciation_dict[key] = [word, replacement]

# pronunciation_dict is a dictionary where the keys are words and the value is a list
# containing the original word and the replacement pronunciation string.
pronunciation_dict: dict[str, list[str]] = {}
